% Mumford-Shah functional minimization through Ambrosio-Tortorelli approximation

clear, clc, close all

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Problem setup
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Image to segment
image = 'butters.jpg';

% Main parameters 
lambda = 10; % Smoothness 
nu = 1e-7; % Discontinuity length
epsilon = 1e-6; % Ambrosio-Tortorelli parameter
maxit = 100; % Max number of iterations
tol = 1e-6; % Stopping criterion (relative energy variation)
thr = .5; % Discontinuity == (1-w)>thr
nu4eps = .25*nu/epsilon;

% Parameters for the inner linear problems if solved with PCG
tol_pcg = 1e-6; % Stopping criterion for conjugate gradient iterations
maxit_pcg = 10; % Stopping criterion for conjugate gradient iterations
LP_pcg = []; % Left preconditioner for conjugate gradient
RP_pcg = []; % Right preconditioner for conjugate gradient

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Auxiliary variables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

I = rgb2gray(double(imread(image))./255); % Image to denoise
[ny,nx,nc] = size(I); % Problem size
[Grad,Dx,Dy] = spmat_gradient2d(ny,nx,nc); % Gradient operators
Id = speye(ny*nx); % Identity matrix
E = NaN*ones(1,maxit+1); % Energy values w.r.t. time

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Initialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Outputs initialization
u = I; % Denoised image
w = ones(ny,nx); % Smooth part indicator function
C = (1-w)>thr; % Discontinuity curve

% Initial energy
E_curr = sum((I(:)-u(:)).^2)... % Data term
	+ lambda*sum((w(:).*(Dx*u(:))).^2)... % Smoothness
	+ lambda*sum((w(:).*(Dy*u(:))).^2)... % Smoothness
	+ nu*epsilon*sum((Dx*w(:)).^2)... % Length
	+ nu*epsilon*sum((Dy*w(:)).^2)... % Length
	+ (0.25*nu/epsilon)*sum((w(:)-1).^2); %Length
E(1) = E_curr;

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Setup displays
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

subplot(2,3,1)
imshow(uint8(255*I))
axis image
title('Noisy image $$I$$','Interpreter','Latex')
subplot(2,3,2)
imshow(uint8(255*u))
axis image
title('Denoised image $$u$$','Interpreter','Latex')
subplot(2,3,4)
imshow(uint8(255*w))
axis image
title('Indicator $$w$$','Interpreter','Latex')
subplot(2,3,5)
imshow(uint8(255*C))
axis image
title('Discontinuity set $$C$$','Interpreter','Latex')

subplot(1,3,3)
semilogy(0:maxit,E,'Linewidth',2)
xlabel('Iterations','Interpreter','Latex')
ylabel('Energy','Interpreter','Latex')
axis([0 maxit 100 E(1)])

drawnow

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Main loop
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for it = 1:maxit % Alternating optimization loop
	
    % Implement the Ambrosio-Tortorelli update step here:
    %
    % 1) Update image u
    % 2) Update discontinuity indicator w
    % 3) Update discontinuity curve C (for visualization purposes)
    % 4) Compute energy
    %
    % For solving the linear systems, you can use the preconditioned 
    % conjugated gradients method (PCG) instead of simple backslash to
    % speed up things.
    % The syntax for solving a system A*x=b using PCG is
    % x = pcg(A, b, tol_pcg, maxit_pcg, LP_pcg, RP_pcg, x0);
    % where x0 is a guess for the solution and the other parameters are
    % specified above in the problem setup.
	
	% Update displays
	subplot(2,3,2)
	imshow(uint8(255*u))
	axis image
	title('Denoised image $$u$$','Interpreter','Latex')
	subplot(2,3,4)
	imshow(uint8(255*w))
	axis image
	title('Indicator $$w$$','Interpreter','Latex')
	subplot(2,3,5)
	imshow(uint8(255*C))
	axis image
	title('Discontinuity set $$C$$','Interpreter','Latex')
	
    subplot(1,3,3)
	semilogy(0:maxit,E,'Linewidth',2)
	xlabel('Iterations','Interpreter','Latex')
	ylabel('Energy','Interpreter','Latex')
    axis([0 maxit 100 E(1)])
	
	drawnow
	
	% Test convergence
    if it > 2 && abs(E(it+1)-E(it))/abs(E(1)) < tol
        disp(['Convergence after ',num2str(it),' iterations!'])
		break
	end
	
end


